<?php
// $Id: operations.node.inc,v 1.3 2009/04/07 14:37:39 jpetso Exp $
/**
 * @file
 * Drupal Transformations -
 * Transformations for processing Drupal-native data like nodes, users, etc.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * Helper class for providing implementations for standard input/output info.
 */
class TfNodeConnectionInfoHelper {
  public static function nodeTypeOptions(&$typeOptionsInfo, $prefix) {
    $typeOptionsInfo[$prefix . 'nodeType']['label'] = t('Node type');
    $typeOptions = array();

    foreach (node_get_types('names') as $type => $name) {
      $typeOptions[$type] = check_plain($name);
    }
    $typeOptionsInfo[$prefix . 'nodeType']['options'] = $typeOptions;
  }

  public static function nodeObjectInfo($propertyKey, $nodeType = NULL) {
    switch ($propertyKey) {
      case 'label':
        return t('Node object');

      case 'expectedSchema':
        $schema = new TfWritableSchemaElement('csvText', 'php:type:string');
        if (isset($nodeType)) {
          $schema->setFormat('drupal:node:object:' . $nodeType);
        }
        return $schema;
    }
  }

  public static function nodeTypeInfo($propertyKey) {
    switch ($propertyKey) {
      case 'label':
        return t('Node type');

      case 'expectedSchema':
        return new TfWritableSchemaElement('node', 'php:type:string', 'drupal:node:type');
    }
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeLoad() {
  return array(
    'label' => t('Load node'),
    'description' => t('Retrieves a Drupal node from the database.'),
  );
}

class TfDrupalNodeLoad extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('nid', 'nodeType');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'nid') {
      switch ($propertyKey) {
        case 'label':
          return t('nid');

        case 'expectedSchema':
          return new TfWritableSchemaElement('nid', 'php:pseudo-type:number');
      }
    }
    elseif ($inputKey == 'nodeType') {
      switch ($propertyKey) {
        case 'required':
          return FALSE;

        case 'description':
          return t('If set, only nodes of the specified type will be loaded. (If the given nid corresponds to a node of a different type, this load operation will fail.) In return, the output schema will be more accurate and contain fields that specifically apply to this node.');

        default:
          return TfNodeConnectionInfoHelper::nodeTypeInfo($propertyKey);
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('node');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'node') {
      return TfNodeConnectionInfoHelper::nodeObjectInfo($propertyKey,
        $this->isInputSet('nodeType') ? $this->input('nodeType')->data() : NULL
      );
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $param = array('nid' => $this->input('nid')->data());
    if ($this->isInputSet('nodeType')) {
      $param['type'] = $this->input('nodeType')->data();
    }
    $node = node_load($param);

    if (!$node) {
      $output->setErrorMessage(t('Could not load node with nid !nid.', array(
        '!nid' => $nid,
      )));
      return;
    }
    $output->set('node', $node);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeSave() {
  return array(
    'label' => t('Save node'),
    'description' => t('Retrieves a Drupal node from the database.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_drupal_operation_TfDrupalNodeSave_type_options(&$typeOptionsInfo, $prefix) {
  TfNodeConnectionInfoHelper::nodeTypeOptions($typeOptionsInfo, $prefix, FALSE);
}

class TfDrupalNodeSave extends TfOperation {
  private $nodeType;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    parent::initialize($typeOptions);
    $this->nodeType = empty($typeOptions['nodeType'])
      ? FALSE
      : $typeOptions['nodeType'];
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    return array('nodeType' => $this->nodeType);
  }

  /**
   * Overriding TfOperation::hasSideEffects() to return TRUE.
   */
  public function hasSideEffects() {
    return TRUE;
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('node');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'node') {
      return TfNodeConnectionInfoHelper::nodeObjectInfo($propertyKey,
        empty($this->nodeType) ? NULL : $this->nodeType
      );
    }
    elseif ($inputKey == 'nodeType') {
      switch ($propertyKey) {
        case 'required':
          return FALSE;

        case 'description':
          return t('If you know the node type of the saved node upfront, you can set this parameter in order to get a more accurate output schema. This save operation will fail if a node type is provided but the node parameter is of a different node type.');

        default:
          return TfNodeConnectionInfoHelper::nodeTypeInfo($propertyKey);
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('savedNode');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'savedNode') {
      switch ($propertyKey) {
        case 'label':
          return t('Saved node object');

        default:
          return TfNodeConnectionInfoHelper::nodeObjectInfo(
            $propertyKey, empty($this->nodeType) ? NULL : $this->nodeType
          );
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $node = $this->input('node')->data();

    if (!empty($this->nodeType) && ($node->type != $this->nodeType)) {
      $output->setErrorMessage(t('Node parameter does not match the given node type !type.', array(
        '!type' => $this->nodeType,
      )));
      return;
    }
    $node = clone $node;
    $node = node_submit($node);
    node_save($node);
    $output->set('savedNode', $node);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeNew() {
  return array(
    'label' => t('Create new node object'),
    'description' => t('Create a barebone Drupal node object. Results are not yet saved to the database.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_drupal_operation_TfDrupalNodeNew_type_options(&$typeOptionsInfo, $prefix) {
  TfNodeConnectionInfoHelper::nodeTypeOptions($typeOptionsInfo, $prefix);
}

class TfDrupalNodeNew extends TfOperation {
  private $nodeType;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    parent::initialize($typeOptions);
    $this->nodeType = empty($typeOptions['nodeType'])
      ? FALSE
      : $typeOptions['nodeType'];
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    return array('nodeType' => $this->nodeType);
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('user');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'user') {
      switch ($propertyKey) {
        case 'label':
          return t('Node author');

        case 'required':
          return FALSE;

        case 'expectedSchema';
          ///TODO: better schema
          return TfWritableSchemaElement('user', 'php:type:object', 'drupal:user:object');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('newNode');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'newNode') {
      switch ($propertyKey) {
        case 'label':
          return t('New node object');

        default:
          return TfNodeConnectionInfoHelper::nodeObjectInfo(
            $propertyKey, $this->nodeType
          );
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $node = new stdClass();
    $node->type = $this->nodeType;

    if ($this->isInputSet('user')) {
      $account = $this->input('user');
      $node->uid = $account->uid;
      $node->name = $account->name;
    }
    else {
      $node->uid = 0;
      $node->name = variable_get('anonymous', t('Anonymous'));
    }
    module_load_include('inc', 'node', 'node.pages');
    node_object_prepare($node);
    $output->set('newNode', $node);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeDelete() {
  return array(
    'label' => t('Delete node'),
    'description' => t('Deletes a Drupal node from the database. Note that it cannot be recovered after it\'s gone!'),
  );
}

class TfDrupalNodeDelete extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('nid');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'nid') {
      switch ($propertyKey) {
        case 'label':
          return t('nid');

        case 'expectedSchema':
          return new TfWritableSchemaElement('nid', 'php:pseudo-type:number');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array();
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    return NULL;
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    node_delete($this->input('nid')->data());
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeFieldsExtract() {
  return array(
    'label' => t('Extract fields from node object'),
    'description' => t('Retrieve data from an existing Drupal node object.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_drupal_operation_TfDrupalNodeFieldsExtract_type_options(&$typeOptionsInfo, $prefix) {
  TfNodeConnectionInfoHelper::nodeTypeOptions($typeOptionsInfo, $prefix, FALSE);
}

class TfDrupalNodeFieldsExtract extends TfOperation {
  private $nodeType;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    parent::initialize($typeOptions);
    $this->nodeType = empty($typeOptions['nodeType'])
      ? FALSE
      : $typeOptions['nodeType'];
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    return array('nodeType' => $this->nodeType);
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('node');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'node') {
      return TfNodeConnectionInfoHelper::nodeObjectInfo($propertyKey,
        empty($this->nodeType) ? NULL : $this->nodeType
      );
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array_keys(fieldtool_fields('node', array(
      'types' => empty($this->nodeType) ? array() : array($this->nodeType),
      'required keys' => array('getter callback'),
    )));
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    $nodeFields = fieldtool_fields('node', array(
      'types' => empty($this->nodeType) ? array() : array($this->nodeType),
      'required keys' => array('getter callback'),
    ));
    return isset($nodeFields[$outputKey][$propertyKey])
      ? $nodeFields[$outputKey][$propertyKey]
      : NULL;
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $node = $this->input('node')->data();
    $nodeFieldInfo = fieldtool_fields('node', array(
      'types' => empty($this->nodeType) ? array() : array($this->nodeType),
      'required keys' => array('getter callback'),
    ));
    foreach ($nodeFieldInfo as $fieldName => &$info) {
      $fieldData = fieldtool_get($info, $node);
      if (isset($fieldData)) {
        $output->set($fieldName, $fieldData);
      }
    }
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_drupal_operation_TfDrupalNodeFieldsSet() {
  return array(
    'label' => t('Set fields in node object'),
    'description' => t('Modify an existing Drupal node object. Results are not yet saved to the database.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_drupal_operation_TfDrupalNodeFieldsSet_type_options(&$typeOptionsInfo, $prefix) {
  TfNodeConnectionInfoHelper::nodeTypeOptions($typeOptionsInfo, $prefix, FALSE);
}

class TfDrupalNodeFieldsSet extends TfOperation {
  private $nodeType;

  /**
   * Overriding TfOperation::initialize().
   */
  public function initialize(Array $typeOptions = array()) {
    parent::initialize($typeOptions);
    $this->nodeType = empty($typeOptions['nodeType'])
      ? FALSE
      : $typeOptions['nodeType'];
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    return array('nodeType' => $this->nodeType);
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    $inputs = array_keys(fieldtool_fields('node', array(
      'types' => empty($this->nodeType) ? array() : array($this->nodeType),
      'required keys' => array('setter callback'),
    )));
    array_unshift($inputs, 'node');
    return $inputs;
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'node') {
      switch ($propertyKey) {
        case 'label':
          return t('Existing node object');

        default:
          return TfNodeConnectionInfoHelper::nodeObjectInfo(
            $propertyKey, empty($this->nodeType) ? NULL : $this->nodeType
          );
      }
    }
    else { // handling field keys
      if ($propertyKey == 'required') {
        return FALSE;
      }
      $nodeFields = fieldtool_fields('node', array(
        'types' => empty($this->nodeType) ? array() : array($this->nodeType),
        'required keys' => array('setter callback'),
      ));
      return isset($nodeFields[$inputKey][$propertyKey])
        ? $nodeFields[$inputKey][$propertyKey]
        : NULL;
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('node');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'node') {
      switch ($propertyKey) {
        case 'expectedSchema':
          return $this->inputProperty('node', 'assignedSchema');

        default:
          return TfNodeConnectionInfoHelper::nodeObjectInfo($propertyKey,
            empty($this->nodeType) ? NULL : $this->nodeType
          );
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $oldNode = $this->input('node')->data();
    $newNode = clone $oldNode;
    $nodeFieldInfo = fieldtool_fields('node', array(
      'types' => empty($this->nodeType) ? array() : array($this->nodeType),
      'required keys' => array('setter callback'),
    ));
    if (isset($nodeFieldInfo['node'])) {
      // Not aware of any module that provides a field named 'node',
      // but let's make sure to avoid those in case there is such a thing.
      unset($nodeFieldInfo['node']);
    }

    foreach ($nodeFieldInfo as $fieldName => &$info) {
      if (!$this->isInputSet($fieldName)) {
        // The user decides which fields to set and which ones to leave as is.
        continue;
      }
      fieldtool_set($info, $newNode, $this->input($fieldName)->data());
    }
    $output->set('node', $newNode);
  }
}
